module plane_search_2_mod

  use norm_mod

  implicit none

  real, dimension (:), allocatable, private   :: s, ss
  integer, private   :: ps_2_iter  ! No. of iteration on plane-search (alpha,beta)
  integer, private   :: ps_2_shrinkthres !How many times the program tries if there is over shot, 1 means half the stepsize once
  integer, private   :: len_datres, len_modres  !length of model residue and data residue
  integer, private   :: len_mod !length of model vector
  real, dimension (:), allocatable, private   :: error_cg ! sum(norm(r)) at each inner iteration, should be a vector of length ps_2_iter.
  real, dimension (:), allocatable, private   :: error_d,error_m! sum(norm(rd)) and sum(norm(rm))at each inner iteration, should be a vector of length ps_2_iter.
  real, dimension (:), allocatable, private   :: avg_rd,avg_rm  ! Not implemented yet
  type(norm_obj), pointer,private :: norm_modres,norm_datres
  real, parameter :: mach_precision = 1e-8         ! relative numerical precision

  double precision, dimension(:), allocatable,private     :: sds, gdg, gds
  double precision, dimension(:), allocatable,private     :: c0,c1,c2
  
  logical, private  :: l_steep_descent ! if true, Use line search only
  
  logical, private  :: lVerb

contains 
  ! ========================================================================================
  !                          Plane Search Initialization
  ! ========================================================================================
  subroutine ps_2_init(psiter_in, ps_2_shrinkthres_in, len_datres_in, len_modres_in, len_mod_in, norm_datres_in, norm_modres_in, line_search_in, verb)
    !psiter_in : Number of Plane-search iterations on Alpha,Beta
    optional  :: line_search_in

    type(norm_obj), target :: norm_modres_in, norm_datres_in
    integer :: psiter_in, ps_2_shrinkthres_in
    integer :: len_datres_in, len_modres_in, len_mod_in
    
    logical  line_search_in, verb

    integer :: len_total_res
    
    l_steep_descent = .false. !Default use plane search
    if (present(line_search_in)) then
       l_steep_descent = line_search_in
    end if

    norm_modres => norm_modres_in
    norm_datres => norm_datres_in

    ps_2_iter = psiter_in
    ps_2_shrinkthres = ps_2_shrinkthres_in

    len_modres = len_modres_in
    len_datres = len_datres_in
    len_mod = len_mod_in

    if (allocated(ss)) then
       deallocate(ss,s)
       deallocate(error_cg,avg_rd,avg_rm)
       deallocate(error_m, error_d)
    end if
    
    len_total_res = len_modres + len_datres
    !Allocate memory
    allocate(ss(len_total_res),s(len_mod))
    allocate (error_cg(ps_2_iter),avg_rd(ps_2_iter),avg_rm(ps_2_iter))
    allocate (error_m(ps_2_iter), error_d(ps_2_iter))
    allocate (c0(len_total_res),c1(len_total_res),c2(len_total_res)) 
    allocate (gdg(len_total_res),sds(len_total_res),gds(len_total_res)) 

	 lVerb = verb
  end subroutine ps_2_init

  ! --------------------------------------------
  !        De-Allocate 
  ! --------------------------------------------
  subroutine ps_2_close( )
    if (allocated(ss)) then
       deallocate(ss,s)
       deallocate(error_cg,avg_rd,avg_rm)
       deallocate(error_d, error_m)
       deallocate(gdg,sds,gds,c0,c1,c2)
    end if
  end subroutine ps_2_close

  ! ========================================================================================
  !                          Compute the both the dat res and mod res 
  ! ========================================================================================
  subroutine ps_2_compute_res_norm(rr_in,c0_out,c1_out,c2_out)
    real, dimension(:), target ::rr_in
    double precision, dimension(:),target ::c0_out
    double precision, dimension(:),target ::c1_out
    double precision, dimension(:),target ::c2_out

    real, dimension(:), pointer :: rd
    real, dimension(:), pointer :: rm
    double precision, dimension(:), pointer :: c0_rd
    double precision, dimension(:), pointer ::  c0_rm
    double precision, dimension(:), pointer ::  c1_rd
    double precision, dimension(:), pointer ::  c1_rm
    double precision, dimension(:), pointer ::  c2_rd
    double precision, dimension(:), pointer ::  c2_rm

    integer ::len_r
    integer ::stat

    len_r = size(rr_in)

    rd => rr_in(1:len_datres)
    c0_rd => c0_out(1:len_datres)
    c1_rd => c1_out(1:len_datres)
    c2_rd => c2_out(1:len_datres)

    !compute norm on data res
    stat = compute_norm(norm_datres, rd, c0_rd, c1_rd, c2_rd)

    if (len_r > len_datres) then !With regularization 
       rm => rr_in(len_datres+1:len_r)
       c0_rm => c0_out(len_datres+1:len_r)
       c1_rm => c1_out(len_datres+1:len_r)
       c2_rm => c2_out(len_datres+1:len_r)
       stat = compute_norm(norm_modres,rm,c0_rm,c1_rm,c2_rm)
    else if (len_r == len_datres) then !Simple solver, no regularization (a.k.a no model residue)
    else
       call erexit('error in compute_res_norm_ps function')
    end if
  end subroutine ps_2_compute_res_norm

  ! ========================================================================================
  !                          Evaluate the change of penalty function
  !   return:  0   success
  !   return:  -1  overshot, norm increase
  !   return:  1   value does not change, should exit plane search
  ! ========================================================================================
  integer function ps_2_evaluate_pf(prev_sum_norm,sum_norm,r,c_alfa_out,c_beta_out,alfa,beta,ps)
    integer :: ps
    double precision:: prev_sum_norm,sum_norm, rat
    real, dimension(:) :: r
    double precision :: c_alfa_out,c_beta_out
    double precision :: alfa,beta
    real :: R_tmp,R_tmp2
    !Consider the numerical precision.
    rat = (prev_sum_norm-sum_norm)/prev_sum_norm   !!It is possible that we have a divide by zero error, think about the case of Operator F being Identity Matrix, we need only 1 iter to converge.
    if ( rat > mach_precision) then
       prev_sum_norm = sum_norm
       
       R_tmp = 0; R_tmp2 = 0
       !call ps_2_compute_avg_r(r,R_tmp,R_tmp2)
       
		 if (lVerb) then
	       write(0,*) "Object Function improved by rat of ", rat
		 end if
       !set all the remaining elements in error to 'sum_norm' in case the loop is exited early
       error_cg(ps-1:) = prev_sum_norm

       error_d(ps-1:)  = sum(c0(1:len_datres))
       error_m(ps-1:)  = sum(c0(len_datres+1:len_datres+len_modres))

       avg_rd(ps-1:) = R_tmp
       avg_rm(ps-1:) = R_tmp2
       !submit alfa,beta update
       c_alfa_out = c_alfa_out+alfa
       c_beta_out = c_beta_out+beta
       ps_2_evaluate_pf = 0

    else if ((prev_sum_norm-sum_norm)/prev_sum_norm > -mach_precision) then ! Here the previous target value is very close to the current one. Which means the search is pretty local now, there is no much space for improvement, exit loop
		 if (lVerb) then
	       write(0,*) 'Plane search ceased, no more improvement!'
		 end if

       ps_2_evaluate_pf = 1 !exit plane search loop
    else  ! Overshot
		 if (lVerb) then
	       write(0,*) 'Plane search overshot, the penalty function increases!!!'
   	    write(0,*) 'sum_norm=',sum_norm,' While prev sum_norm=', prev_sum_norm
		 end if
       ps_2_evaluate_pf = -1
    endif
  end function ps_2_evaluate_pf

  ! ========================================================================================
  !                          Compute  both the avg_rd and avg_rm 
  ! ========================================================================================
  subroutine ps_2_compute_avg_r(rr_in,avg_rd_out, avg_rm_out)
    real, dimension(:),target :: rr_in
    real :: avg_rd_out, avg_rm_out

    real, dimension(:), pointer :: rd
    real, dimension(:), pointer :: rm
    integer ::len_r
    integer ::stat

    len_r = size(rr_in)
    rd => rr_in(1:len_datres)

    stat = compute_avg_r(norm_datres, rd, avg_rd_out)

    avg_rm_out = 0
    if (len_r > len_datres) then ! Need to compute avg_rm
       rm => rr_in(len_datres+1:len_r)
       stat = compute_avg_r(norm_modres, rm, avg_rm_out)
    end if
  end subroutine ps_2_compute_avg_r

  ! --------------------------------------------
  !              Plane Search function Yang
  ! --------------------------------------------
  integer function cgstep_ps_2( forget, x, g, rr, gg)

    ! Iterative Plane-Search Stepper function 
    !    for different norms (Hybrid, Huber, etc.)

    ! Arguments: 
    ! x, g   : m, dm
    ! rr,gg  : R, dR
    ! forget : set steepest-decsend if true and CG-direction if false

    ! Output Flag (cgnorm): 
    !   2 : Exit
    !   1 : Exit with vanishing gradient status
    !   0 : Done without any exit or error
    !


    real, dimension (:)  :: x, g, rr, gg
    logical              :: forget 
    double precision     :: alfa, beta

    integer              :: i,ps,stat,stat2
    integer			   :: nr  ! length of residue vector
    integer			   :: nx  ! length of model vector
    double precision	   :: prev_sum_norm  !previous value of penalty function, e.g  sum of norm(r)	
    double precision	   :: sum_norm

    integer :: ps_2_shrinkcnt
    double precision	   :: c_alfa,c_beta !Cumulative alfa & beta


    real,dimension (:), allocatable :: tmp_ss ! Temporary vector stores solution_step and residue step
    real,dimension (:), allocatable :: tmp_rr ! Temporary vector stores residue
    real :: R_tmp, R_tmp2
    logical :: L_quit

    cgstep_ps_2 = 0  ! preset the return value of this function, 0 means successful

    prev_sum_norm = 0.0
    sum_norm = 0.0
    nr = size(rr)
    nx = size(x)

    if (nr.ne.(len_modres+len_datres) .or. nx.ne.len_mod) then
       call erexit('vector length does not match in cgstep_ps_2 func, exit!!')
    endif

    !Allocate tmp array in every invoke, this might casue performance degrade.
    allocate (tmp_ss(nr),tmp_rr(nr))

    tmp_rr = rr
    
    error_cg=0

    avg_rm = 0; avg_rd = 0;

    if(forget == .true.) then
       !clean the memory of previous step
       s  = 0
       ss = 0
    end if
    tmp_ss = ss

    c_alfa=0.0; c_beta=0.0
    alfa  =0.0; beta  =0.0

    L_quit = .false.
    !write(0,*) "Enter the plane search loop in ZY"
    ! Loop over alpha , Beta
    do ps=1,ps_2_iter+1  !add one more loop to compute the ps_2_norm ps_2_iter+1 times, since we need the last time to report the error amplitude

       ! === Compute Norm and its derivatives
       !obs !stat = norm_modulus(norm_datres,tmp_rr, R_tmp)
       call ps_2_compute_res_norm(tmp_rr,c0,c1,c2)

       if (ps==1) then
          prev_sum_norm = sum(c0)
          error_cg(1:) = prev_sum_norm !Initalize the error vector to be the current one
          error_d(1:)  = sum(c0(1:len_datres))
          error_m(1:)  = sum(c0(len_datres+1:len_datres+len_modres))

          R_tmp=0; R_tmp2=0
          !call ps_2_compute_avg_r(tmp_rr,R_tmp,R_tmp2)
          avg_rd(1:) = R_tmp
          avg_rm(1:) = R_tmp2
       else
          !!compute new penalty fun (sum(norm)), make sure the function value is descending
          sum_norm = sum(c0)
          stat = ps_2_evaluate_pf(prev_sum_norm,sum_norm,tmp_rr,c_alfa,c_beta,alfa,beta,ps)

          if (sum_norm < mach_precision * 100)  then ! This case considered that it already converges
             cgstep_ps_2 = 9
             exit
          end if
          if (stat==1) then!Exit plane search 
             exit
          end if
          if (stat==-1) then ! Overshot
             ! Strategy 1, half the step size to see it works or not
             do ps_2_shrinkcnt=1,ps_2_shrinkthres,1
                !Retry with the halved step size
                alfa = alfa*0.5; beta = beta*0.5
                tmp_ss	= alfa*gg + beta*ss
                tmp_rr = rr  + c_alfa*gg + c_beta*ss + tmp_ss
                !evaluate the norm again
                call ps_2_compute_res_norm(tmp_rr,c0,c1,c2)
                sum_norm = sum(c0)
                stat2 = ps_2_evaluate_pf(prev_sum_norm,sum_norm,tmp_rr,c_alfa,c_beta,alfa,beta,ps)
                if (stat2==1) then
                   L_quit = .true.
                   exit !!!!! We should exit outer loop
                else if (stat2==-1)  then !still over shot
                   cycle
                else ! (stat2==0) success
                   exit !exit this do loop
                end if
             end do !ps_2_shrinkcnt

             if (ps_2_shrinkcnt == ps_2_shrinkthres+1) then
					 if (lVerb) then
						 write(0,*) 'Plane search failed, reaching the maximum trial steps, alfa',alfa,'beta',beta
					 end if
                L_quit =.true.
             end if

             if (L_quit == .true.) then
                exit
             end if

          end if
       end if

       if (ps==ps_2_iter+1)  then !we don't need to execute the code reamaining in the loop for the last time
          exit
       end if


       !compute alfa and beta
       stat = solve_alfa_beta(forget,gg,alfa,beta)
       if (stat .ne. 0) then 
          !Something wrong happens, quit the loop
          write(0,*) "Something wrong happens in func solve_alfa_beta"
          exit
       end if

       tmp_ss	= alfa*gg + beta*ss
       tmp_rr = tmp_rr  + tmp_ss
       !write (0,*) "Alpha, Beta::" ,alfa,beta
       !write (0,*) "Determ::" ,determ
    end do !End of the iteration loop

    if (.not. abs(c_alfa)<1e-15) then
       if (forget == .true.) then
          s  = c_alfa * g
          ss = c_alfa * gg
          if (.not. l_steep_descent) then
             forget = .false.
          end if
       else
          s  = c_alfa *  g + c_beta *  s            ! update solution step
          ss = c_alfa * gg + c_beta * ss            ! update residual step
       end if
       x  =  x +  s                           ! update solution
       rr = rr + ss	 			! update residual
    else !If even the first search fails, c_alfa=0, we do not change anything, but we tell the caller that no more iterations in the outerloop is needed
       write(0,*) 'c_alfa=',c_alfa
       cgstep_ps_2 = 1
    end if

    deallocate (tmp_ss,tmp_rr)

    !write(0,*) "Finish the plane search loop in ZY"
  end function cgstep_ps_2



  ! ========================================================================================
  !                   solve_alfa_beta routine
  ! ========================================================================================
  !!  solve the 2by2 equation to get alfa and beta, compute alfa only if forget==true, use deepest descent.
  !!  Arguments OUT: 
  !!     alfa,beta
  !!---------------------------------------
  !! return 0: success

  integer function  solve_alfa_beta(forget,gg,alfa,beta)
    double precision,intent(out)   :: alfa, beta
    double precision               :: determ,c1dg,c1ds,c2dgg,c2dss,c2dgs
    real :: sin0_2       !sin(theta)**2, indicate the orthoganlity of search directions

    real, dimension(:)  :: gg
    logical, intent(in) :: forget

    solve_alfa_beta = 0

    ! === Pick Line Search or Plane Search
    if( forget) then
       !write(0,*) "enter line search, deepest descent"
       ! Steepest Descent
       beta = 0.0

       gdg   = dprod(gg,gg)
       c2dgg = sum(dprod(c2,gdg))

       if( c2dgg == 0.0 )  then
	solve_alfa_beta = 1
          call erexit('NORM_CGSTEP: Gradient vanishes identically: C"*gg*gg=0 .')
       end if

       alfa = - sum( dprod(c1,gg)) / c2dgg
    else
       ! Plane Search for alpha & beta
       !write(0,*) "enter the plane search"
       c1dg = - sum(dprod(c1,gg))   ! c1dg = -C'*G
       c1ds = - sum(dprod(c1,ss))   ! c1ds = -C'*S

       gdg = dprod(gg,gg)           ! search plane by solving 2-by-2
       sds = dprod(ss,ss)           ! C"*G . (G*alfa + S*beta) = -C'*G
       gds = dprod(gg,ss)           ! C"*S . (G*alfa + S*beta) = -C'*S

       c2dgg = sum(dprod(c2,gdg))   ! c2dgg = C"*G*G
       c2dss = sum(dprod(c2,sds))   ! c2dss = C"*S*S
       c2dgs = sum(dprod(c2,gds))   ! c2dgs = C"*G*S

       if ((c2dgg==0.d0) .OR. (c2dss==0.d0)) then
          !Since C2 is always greater than zero, this case should not happen
          write (0,*) 'Plane Search : Gradient vanishes identically: C"*gg*gg=0 OR C"*ss*ss=0., c2dgg,c2dss=',c2dgg,c2dss
          solve_alfa_beta = 1
          !call erexit('Plane Search : Gradient vanishes identically')
          return
       end if

       !write(0,*) "bivariate search, sum(gds), sum(gdg), sum(sds) =",sum(gds),sum(gdg),sum(sds)

       ! Here compute the extent of orthogonality
       ! cos(theta) = (AdotB)/(|A||B|)
       sin0_2 = 1- (sum(gds)/sum(gdg))*(sum(gds)/sum(sds))
       if (sin0_2 < 1e-3) then
          write(0,*) "g and s are highly linear dependent! plane search ceased!"
          solve_alfa_beta = 1
       end if
       
       
       determ = 1.d0 - (c2dgs/c2dgg)*(c2dgs/c2dss)
       !write (0,*) "the determ scale is=", determ, "sin0^2=", sin0_2

       determ = c2dgg * c2dss * determ

       if (determ < 1.d-12) then
          write(0,*) "c2dgg, c2dss, sin^2=", c2dgg, c2dss, determ
          write(0,*) "the determinant of C_2 matrix is too small! plane search ceased"
          solve_alfa_beta = 1
       end if

       alfa = ( c2dss * c1dg - c2dgs * c1ds ) / determ
       beta = (-c2dgs * c1dg + c2dgg * c1ds ) / determ

    end if
  end function solve_alfa_beta

  ! ========================================================================================
  !                   QueryError Functions
  ! ========================================================================================
  !!  Output the fitting error at each iteration 
  !!  Arguments OUT: 
  !!     err_out:a real array of size ps_2_iter at least
  !!---------------------------------------
  !! 
  subroutine ps_2_QueryError(err_out,err_rd_out,avg_rd_out,err_rm_out,avg_rm_out)
    optional :: avg_rm_out, err_rm_out, err_rd_out
    real, dimension (:)  :: err_out, err_rd_out, avg_rd_out, err_rm_out, avg_rm_out

    if (allocated(s)) then
		 if (lVerb) then
       	write(0,*) "error in plane search = ",error_cg
		end if
       err_out(1:ps_2_iter) = error_cg
       avg_rd_out(1:ps_2_iter) = avg_rd
       if (present(avg_rm_out)) then
          avg_rm_out(1:ps_2_iter) = avg_rm
       end if
       if (present(err_rm_out)) then
          err_rm_out(1:ps_2_iter) = error_m
       end if
      if (present(err_rd_out)) then
          err_rd_out(1:ps_2_iter) = error_d
       end if
    else
       write(0,*) 'ERROR in QueryError! module not initialized yet.'
       stop
    end if
  end subroutine ps_2_QueryError
end module plane_search_2_mod
